"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

conversion_plus.py
~~~~~~~~~~~~~~~~~~

Conversion plus technology constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import \
    get_param, \
    split_comma_list, \
    get_conversion_plus_io


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data__['sets']

    backend_model.balance_conversion_plus_primary_constraint = po.Constraint(
        backend_model.loc_techs_balance_conversion_plus_primary_constraint,
        backend_model.timesteps,
        rule=balance_conversion_plus_primary_constraint_rule
    )

    if 'loc_techs_carrier_production_max_conversion_plus_constraint' in sets:
        backend_model.carrier_production_max_conversion_plus_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_max_conversion_plus_constraint,
            backend_model.timesteps,
            rule=carrier_production_max_conversion_plus_constraint_rule
        )

    if 'loc_techs_carrier_production_min_conversion_plus_constraint' in sets:
        backend_model.carrier_production_min_conversion_plus_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_min_conversion_plus_constraint,
            backend_model.timesteps,
            rule=carrier_production_min_conversion_plus_constraint_rule
        )

    if 'loc_techs_cost_var_conversion_plus_constraint' in sets:
        backend_model.cost_var_conversion_plus_constraint = po.Constraint(
            backend_model.costs, backend_model.loc_techs_cost_var_conversion_plus_constraint,
            backend_model.timesteps,
            rule=cost_var_conversion_plus_constraint_rule
        )

    if 'loc_techs_balance_conversion_plus_in_2_constraint' in sets:
        backend_model.balance_conversion_plus_in_2_constraint = po.Constraint(
            ['in_2'], backend_model.loc_techs_balance_conversion_plus_in_2_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule
        )

    if 'loc_techs_balance_conversion_plus_in_3_constraint' in sets:
        backend_model.balance_conversion_plus_in_3_constraint = po.Constraint(
            ['in_3'], backend_model.loc_techs_balance_conversion_plus_in_3_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule
        )

    if 'loc_techs_balance_conversion_plus_out_2_constraint' in sets:
        backend_model.balance_conversion_plus_out_2_constraint = po.Constraint(
            ['out_2'], backend_model.loc_techs_balance_conversion_plus_out_2_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule
        )

    if 'loc_techs_balance_conversion_plus_out_3_constraint' in sets:
        backend_model.balance_conversion_plus_out_3_constraint = po.Constraint(
            ['out_3'], backend_model.loc_techs_balance_conversion_plus_out_3_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule
        )


def balance_conversion_plus_primary_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance energy carrier consumption and production for carrier_in and carrier_out

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\_ratio(loc::tech::carrier, `out')} =
            -1 * \\sum_{loc::tech::carrier \\in loc::tech::carriers_{in}} (
            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            * carrier\_ratio(loc::tech::carrier, `in') * \\eta_{energy}(loc::tech, timestep))
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    loc_tech_carriers_out = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus']['out', loc_tech]
    )
    loc_tech_carriers_in = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus']['in', loc_tech]
    )

    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))

    carrier_prod = sum(backend_model.carrier_prod[loc_tech_carrier, timestep]
                 / get_param(backend_model, 'carrier_ratios', ('out', loc_tech_carrier))
                 for loc_tech_carrier in loc_tech_carriers_out)
    carrier_con = sum(backend_model.carrier_con[loc_tech_carrier, timestep]
                 * get_param(backend_model, 'carrier_ratios', ('in', loc_tech_carrier))
                 for loc_tech_carrier in loc_tech_carriers_in)

    return carrier_prod == -1 * carrier_con * energy_eff


def carrier_production_max_conversion_plus_constraint_rule(backend_model, loc_tech, timestep):
    """
    Set maximum conversion_plus carrier production.

    .. container:: scrolling-wrapper

        .. math::

            \sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq \\boldsymbol{energy_{cap}}(loc::tech) \\times timestep\_resolution(timestep)
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    timestep_resolution = backend_model.timestep_resolution[timestep]
    loc_tech_carriers_out = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus']['out', loc_tech]
    )

    carrier_prod = sum(backend_model.carrier_prod[loc_tech_carrier, timestep]
                 for loc_tech_carrier in loc_tech_carriers_out)

    return carrier_prod <= timestep_resolution * backend_model.energy_cap[loc_tech]


def carrier_production_min_conversion_plus_constraint_rule(backend_model, loc_tech, timestep):
    """
    Set minimum conversion_plus carrier production.

    .. container:: scrolling-wrapper

        .. math::

            \sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq \\boldsymbol{energy_{cap}}(loc::tech) \\times timestep\_resolution(timestep)
            \\times energy_{cap, min use}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    timestep_resolution = backend_model.timestep_resolution[timestep]
    min_use = get_param(backend_model, 'energy_cap_min_use', (loc_tech, timestep))

    loc_tech_carriers_out = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus']['out', loc_tech]
    )

    carrier_prod = sum(backend_model.carrier_prod[loc_tech_carrier, timestep]
                for loc_tech_carrier in loc_tech_carriers_out)

    return carrier_prod >= (
        timestep_resolution * backend_model.energy_cap[loc_tech] * min_use
    )


def cost_var_conversion_plus_constraint_rule(backend_model, cost, loc_tech, timestep):
    """
    Add time-varying conversion_plus technology costs

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(loc::tech, cost, timestep) =
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier_{primary}, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, prod}(loc::tech, cost, timestep)
            +
            \\boldsymbol{carrier_{con}}(loc::tech::carrier_{primary}, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, con}(loc::tech, cost, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{cost_{var}, conversion^{+}}
    """
    model_data_dict = backend_model.__calliope_model_data__['data']
    weight = backend_model.timestep_weights[timestep]

    loc_tech_carrier = (
        model_data_dict['lookup_primary_loc_tech_carriers'][loc_tech]
    )

    var_cost = 0

    if loc_tech_carrier in backend_model.loc_tech_carriers_prod:
        cost_om_prod = get_param(backend_model, 'cost_om_prod',
                                 (cost, loc_tech, timestep))
        if cost_om_prod:
            var_cost += (
                cost_om_prod * weight *
                backend_model.carrier_prod[loc_tech_carrier, timestep]
            )

    if loc_tech_carrier in backend_model.loc_tech_carriers_con:
        cost_om_con = get_param(backend_model, 'cost_om_con',
                                (cost, loc_tech, timestep))
        if cost_om_con:
            var_cost += (
                cost_om_con * weight *
                backend_model.carrier_con[loc_tech_carrier, timestep]
            )

    backend_model.cost_var_rhs[cost, loc_tech, timestep] = var_cost
    return (backend_model.cost_var[cost, loc_tech, timestep] ==
            backend_model.cost_var_rhs[cost, loc_tech, timestep])


def balance_conversion_plus_tiers_constraint_rule(backend_model, tier, loc_tech, timestep):
    """
    Force all carrier_in_2/carrier_in_3 and carrier_out_2/carrier_out_3 to follow
    carrier_in and carrier_out (respectively).

    If `tier` in ['out_2', 'out_3']:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}} (
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\_ratio(loc::tech::carrier, `out')} =
            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{tier}} (
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\_ratio(loc::tech::carrier, tier)}
            \\quad \\forall \\text { tier } \\in [`out_2', `out_3'], \\forall loc::tech
                \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps

    If `tier` in ['in_2', 'in_3']:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{in}}
            \\frac{\\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)}{
                carrier\_ratio(loc::tech::carrier, `in')} =
            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{tier}}
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\_ratio(loc::tech::carrier, tier)}
            \\quad \\forall \\text{ tier } \\in [`in_2', `in_3'], \\forall loc::tech
                \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps
    """
    primary_tier, decision_variable = get_conversion_plus_io(backend_model, tier)
    model_data_dict = backend_model.__calliope_model_data__['data']

    loc_tech_carriers_1 = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus'][primary_tier, loc_tech]
    )
    loc_tech_carriers_2 = split_comma_list(
        model_data_dict['lookup_loc_techs_conversion_plus'][tier, loc_tech]
    )

    c_1 = sum(decision_variable[loc_tech_carrier, timestep]
        / get_param(backend_model, 'carrier_ratios', (primary_tier, loc_tech_carrier))
        for loc_tech_carrier in loc_tech_carriers_1)
    c_2 = sum(decision_variable[loc_tech_carrier, timestep]
        / get_param(backend_model, 'carrier_ratios', (tier, loc_tech_carrier))
        for loc_tech_carrier in loc_tech_carriers_2)

    return c_1 == c_2



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

from calliope import exceptions

import xarray as xr
import pandas as pd


def get_loc_techs(loc_techs, tech=None, loc=None):
    """
    Get a list of loc_techs associated with the given technology and/or location.
    If multiple of both loc and tech are given, the function will return any
    combination of members of loc and tech lists found in loc_techs.

    Parameters
    ----------
    loc_techs : list
        set of loc_techs to search for the relevant tech and/or loc
    tech : string or list of strings, default None
        technology/technologies to search for in the set of location:technology
    loc : string or list of strings, default None
        location(s) to search for in the set of location:technology

    Returns
    -------
    relevant_loc_techs : list of strings

    """
    # If both are strings, there is only one loc:tech possibility to look for
    if (isinstance(tech, str) and isinstance(loc, str)
        and '::'.join([loc, tech]) in loc_techs):
        relevant_loc_techs = ['::'.join([loc, tech])]

    tech = [tech] if tech is not None and isinstance(tech, str) else tech
    loc = [loc] if loc is not None and isinstance(loc, str) else loc

    if tech and not loc:
        relevant_loc_techs = [i for i in loc_techs if i.split('::')[1] in tech]
    elif loc and not tech:
        relevant_loc_techs = [i for i in loc_techs if i.split('::')[0] in loc]
    elif loc and tech:
        loc_techs_set = set(tuple(i.split('::')) for i in loc_techs)
        possible_loc_techs = set((l, t) for l in loc for t in tech)
        relevant_loc_techs = [
            '::'.join(i) for i in possible_loc_techs.intersection(loc_techs_set)
        ]
    else:
        relevant_loc_techs = [None]

    return relevant_loc_techs


def split_loc_techs(data_var, as_='DataArray'):
    """
    Get a DataArray with locations technologies, and possibly carriers
    split into separate coordinates.

    Parameters
    ----------
    data_var : xarray DataArray
        Variable from Calliope model_data, to split loc_techs dimension
    as_ : string
        'DataArray' to return xarray DataArray or 'Series' to return pandas
        Series with dimensions as a MultiIndex

    Returns
    -------
    updated_data_var : xarray DataArray of pandas Series
    """

    # Separately find the loc_techs(_carriers) dimension and all other dimensions
    loc_tech_dim = [i for i in data_var.dims if 'loc_tech' in i]
    if not loc_tech_dim:
        loc_tech_dim = [i for i in data_var.dims if 'loc_carrier' in i]
    non_loc_tech_dims = list(set(data_var.dims).difference(loc_tech_dim))

    if not loc_tech_dim:
        if as_ == 'Series':
            return data_var.to_series()
        elif as_ == 'DataArray':
            return data_var
        else:
            raise ValueError('`as_` must be `DataArray` or `Series`, '
                             'but `{}` given'.format(as_))

    elif len(loc_tech_dim) > 1:
        e = exceptions.ModelError
        raise e("Cannot split loc_techs or loc_techs_carrier dimension "
                "for DataArray {}".format(data_var.name))

    loc_tech_dim = loc_tech_dim[0]
    # xr.Datarray -> pd.Series allows for string operations
    data_var_df = data_var.to_series().unstack(non_loc_tech_dims)
    index_list = data_var_df.index.str.split('::').tolist()

    # carrier_prod, carrier_con, and carrier_export will return an index_list
    # of size 3, all others will be an index list of size 2
    possible_names = ['loc', 'tech', 'carrier']
    names = [i + 's' for i in possible_names if i in loc_tech_dim]

    data_var_df.index = pd.MultiIndex.from_tuples(index_list, names=names)

    # If there were no other dimensions other than loc_techs(_carriers) then
    # nothing was unstacked on creating data_var_df, so nothing is stacked now
    if isinstance(data_var_df, pd.Series):
        data_var_series = data_var_df
    else:
        data_var_series = data_var_df.stack(non_loc_tech_dims)

    if as_ == "Series":
        return data_var_series

    elif as_ == "DataArray":
        updated_data_var = xr.DataArray.from_series(data_var_series)
        updated_data_var.attrs = data_var.attrs

        return updated_data_var

    else:
        raise ValueError('`as_` must be `DataArray` or `Series`, '
                         'but `{}` given'.format(as_))


def reorganise_dataset_dimensions(dataset):
    """
    Reorganise the Dataset dimensions to be alphabetical *except*
    `timesteps`, which must always come last in any DataArray's dimensions
    """

    new_dims = (
        sorted(list(set(dataset.dims.keys()) - set(['timesteps'])))
    ) + ['timesteps']

    updated_dataset = dataset.transpose(*new_dims).reindex(
        {k:dataset[k] for k in new_dims})

    return updated_dataset



import os

import pytest  # pylint: disable=unused-import
import tempfile

import calliope


class TestModel:
    def test_save_commented_model_yaml(self):
        model = calliope.examples.national_scale()
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model_debug.yaml')
            model.save_commented_model_yaml(out_path)
            assert os.path.isfile(out_path)



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_checks.py
~~~~~~~~~~~~~~~~~~~~

Checks for model consistency and possible errors during preprocessing.

"""

import os
import logging

import numpy as np
import xarray as xr

from inspect import signature

import calliope
from calliope._version import __version__
from calliope.core.attrdict import AttrDict
from calliope.core.util.tools import flatten_list
from calliope.core.preprocess.util import get_all_carriers
from calliope.core.util.logging import logger
from calliope.core.util.tools import load_function

_defaults_files = {
    k: os.path.join(os.path.dirname(calliope.__file__), 'config', k + '.yaml')
    for k in ['model', 'defaults']
}
defaults = AttrDict.from_yaml(_defaults_files['defaults'])
defaults_model = AttrDict.from_yaml(_defaults_files['model'])


def check_overrides(config_model, override):
    """
    Perform checks on the override dict and override file inputs to ensure they
    are not doing something silly.
    """
    warnings = []
    info = []
    for key in override.as_dict_flat().keys():
        if key in config_model.as_dict_flat().keys():
            info.append(
                'Override applied to {}: {} -> {}'
                .format(key, config_model.get_key(key), override.get_key(key))
            )
        else:
            info.append(
                '`{}`:{} applied from override as new configuration'
                .format(key, override.get_key(key))
            )

    # Check if overriding coordinates are in the same coordinate system. If not,
    # delete all incumbent coordinates, ready for the new coordinates to come in
    if (any(['coordinates' in k for k in config_model.as_dict_flat().keys()]) and
            any(['coordinates' in k for k in override.as_dict_flat().keys()])):

        # get keys that might be deleted and incumbent coordinate system
        config_keys = [k for k in config_model.as_dict_flat().keys() if 'coordinates.' in k]
        config_coordinates = set([k.split('coordinates.')[-1] for k in config_keys])

        # get overriding coordinate system
        override_coordinates = set(
            k.split('coordinates.')[-1] for k in override.as_dict_flat().keys()
            if 'coordinates.' in k
        )

        # compare overriding and incumbent, deleting incumbent if overriding is different
        if config_coordinates != override_coordinates:
            for key in config_keys:
                config_model.del_key(key)
            warnings.append(
                'Updated from coordinate system {} to {}, using overrides'
                .format(config_coordinates, override_coordinates)
            )

    if info:
        logger.info('\n'.join(info))

    return warnings


def check_initial(config_model):
    """
    Perform initial checks of model and run config dicts.

    Returns
    -------
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    errors = []
    warnings = []

    # Check for version mismatch
    model_version = config_model.model.get('calliope_version', False)
    if model_version:
        if not str(model_version) in __version__:
            warnings.append(
                'Model configuration specifies calliope_version={}, '
                'but you are running {}. Proceed with caution!'.format(
                    model_version, __version__)
            )

    # Check top-level keys
    for k in config_model.keys():
        if k not in ['model', 'run', 'locations', 'tech_groups', 'techs', 'links', 'config_path']:
            warnings.append(
                'Unrecognised top-level configuration item: {}'.format(k)
            )

    # Check run configuration
    # Exclude solver_options from checks, as we don't know all possible
    # options for all solvers
    for k in config_model['run'].keys_nested():
        if (k not in defaults_model['run'].keys_nested() and
                'solver_options' not in k):
            warnings.append(
                'Unrecognised setting in run configuration: {}'.format(k)
            )

    # Check model configuration, but top-level keys only
    for k in config_model['model'].keys():
        if k not in defaults_model['model'].keys():
            warnings.append(
                'Unrecognised setting in model configuration: {}'.format(k)
            )

    # Only ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3']
    # are allowed as carrier tiers
    for key in config_model.as_dict_flat().keys():
        if ('.carrier_' in key and key.split('.carrier_')[-1].split('.')[0] not
                in ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3', 'ratios'] and
                'group_share' not in key):
            errors.append(
                "Invalid carrier tier found at {}. Only "
                "'carrier_' + ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3'] "
                "is valid.".format(key)
            )

    # No techs may have the same identifier as a tech_group
    name_overlap = (
        set(config_model.tech_groups.keys()) &
        set(config_model.techs.keys())
    )
    if name_overlap:
        errors.append(
            'tech_groups and techs with '
            'the same name exist: {}'.format(name_overlap)
        )

    # Checks for techs and tech_groups:
    # * All user-defined tech and tech_groups must specify a parent
    # * techs cannot be parents, only tech groups can
    # * No carrier may be called 'resource'
    default_tech_groups = list(defaults_model.tech_groups.keys())
    for tg_name, tg_config in config_model.tech_groups.items():
        if tg_name in default_tech_groups:
            continue
        if not tg_config.get_key('essentials.parent'):
            errors.append(
                'tech_group {} does not define '
                '`essentials.parent`'.format(tg_name)
            )
        elif tg_config.get_key('essentials.parent') in config_model.techs.keys():
            errors.append(
                'tech_group `{}` has a tech as a parent, only another tech_group '
                'is allowed'.format(tg_name)
            )
        if 'resource' in get_all_carriers(tg_config.essentials):
            errors.append(
                'No carrier called `resource` may '
                'be defined (tech_group: {})'.format(tg_name)
            )

    for t_name, t_config in config_model.techs.items():
        if not t_config.get_key('essentials.parent'):
            errors.append(
                'tech {} does not define '
                '`essentials.parent`'.format(t_name)
            )
        elif t_config.get_key('essentials.parent') in config_model.techs.keys():
            errors.append(
                'tech `{}` has another tech as a parent, only a tech_group '
                'is allowed'.format(tg_name)
            )
        if 'resource' in get_all_carriers(t_config.essentials):
            errors.append(
                'No carrier called `resource` may '
                'be defined (tech: {})'.format(t_name)
            )

    # Error if a constraint is loaded from file that must not be
    allowed_from_file = defaults['file_allowed']
    for k, v in config_model.as_dict_flat().items():
        if 'file=' in str(v):
            constraint_name = k.split('.')[-1]
            if constraint_name not in allowed_from_file:
                errors.append(
                    'Cannot load `{}` from file for configuration {}'
                    .format(constraint_name, k)
                )

    # Check the objective function being used has all the appropriate
    # options set in objective_options, and that no options are unused
    objective_function = 'calliope.backend.pyomo.objective.' + config_model.run.objective
    objective_args_expected = list(signature(load_function(objective_function)).parameters.keys())
    objective_args_expected = [arg for arg in objective_args_expected
                               if arg not in ['backend_model', 'kwargs']]
    for arg in objective_args_expected:
        if arg not in config_model.run.objective_options:
            errors.append(
                'Objective function argument `{}` not found in run.objective_options'
                .format(arg)
            )
    for arg in config_model.run.objective_options:
        if arg not in objective_args_expected:
            warnings.append(
                'Objective function argument `{}` given but not used by objective function `{}`'
                .format(arg, config_model.run.objective)
            )

    return warnings, errors


def _check_tech(model_run, tech_id, tech_config, loc_id, warnings, errors, comments):
    """
    Checks individual tech/tech groups at specific locations.
    NOTE: Updates `warnings` and `errors` lists in-place.
    """
    if tech_id not in model_run.techs:
        warnings.append(
            'Tech {} was removed by setting ``exists: False`` - not checking '
            'the consistency of its constraints at location {}.'.format(tech_id, loc_id)
        )
        return warnings, errors

    required = model_run.techs[tech_id].required_constraints
    allowed = model_run.techs[tech_id].allowed_constraints
    allowed_costs = model_run.techs[tech_id].allowed_costs
    all_defaults = list(defaults.default_tech.constraints.keys())

    # Error if required constraints are not defined
    for r in required:
        # If it's a string, it must be defined
        single_ok = isinstance(r, str) and r in tech_config.constraints
        # If it's a list of strings, one of them must be defined
        multiple_ok = (
            isinstance(r, list) and
            any([i in tech_config.constraints for i in r])
        )
        if not single_ok and not multiple_ok:
            errors.append(
                '`{}` at `{}` fails to define '
                'all required constraints: {}'.format(tech_id, loc_id, required)
            )
            # print('{} -- {}-{}: {}, {}'.format(r, loc_id, tech_id, single_ok, multiple_ok))

    # If the technology is supply_plus, check if it has storage_cap_max. If yes, it needs charge rate
    if model_run.techs[tech_id].essentials.parent == 'supply_plus':
        if (any(['storage_cap_' in k for k in tech_config.constraints.keys()])
            and 'charge_rate' not in tech_config.constraints.keys()):
            errors.append(
                '`{}` at `{}` fails to define '
                'charge_rate, but is using storage'.format(tech_id, loc_id, required)
            )
    # If a technology is defined by units (i.e. integer decision variable), it must define energy_cap_per_unit
    if (any(['units_' in k for k in tech_config.constraints.keys()])
        and 'energy_cap_per_unit' not in tech_config.constraints.keys()):
        errors.append(
            '`{}` at `{}` fails to define energy_cap_per_unit when specifying '
            'technology in units_max/min/equals'.format(tech_id, loc_id, required)
        )

    # If a technology is defined by units & is a storage tech, it must define storage_cap_per_unit
    if (any(['units_' in k for k in tech_config.constraints.keys()])
            and model_run.techs[tech_id].essentials.parent in ['storage', 'supply_plus']
            and any(['storage' in k for k in tech_config.constraints.keys()])
            and 'storage_cap_per_unit' not in tech_config.constraints.keys()):
        errors.append(
            '`{}` at `{}` fails to define storage_cap_per_unit when specifying '
            'technology in units_max/min/equals'.format(tech_id, loc_id, required)
        )

    # If a technology is defines force_resource but is not in loc_techs_finite_resource
    if ('force_resource' in tech_config.constraints.keys() and
            loc_id + '::' + tech_id not in model_run.sets.loc_techs_finite_resource):

        warnings.append(
            '`{}` at `{}` defines force_resource but not a finite resource, so '
            'force_resource will not be applied'.format(tech_id, loc_id)
        )

    # Flatten required list and gather remaining unallowed constraints
    required_f = flatten_list(required)
    remaining = set(tech_config.constraints) - set(required_f) - set(allowed)

    # Error if something is defined that's not allowed, but is in defaults
    # Warn if something is defined that's not allowed, but is not in defaults
    # (it could be a misspelling)
    for k in remaining:
        if k in all_defaults:
            errors.append(
                '`{}` at `{}` defines non-allowed '
                'constraint `{}`'.format(tech_id, loc_id, k)
            )
        else:
            warnings.append(
                '`{}` at `{}` defines unrecognised '
                'constraint `{}` - possibly a misspelling?'.format(tech_id, loc_id, k)
            )

    # Error if an `export` statement does not match the given carrier_outs
    if 'export_carrier' in tech_config.constraints:
        essentials = model_run.techs[tech_id].essentials
        export = tech_config.constraints.export_carrier
        if (export and export not in [essentials.get_key(k, '')
                for k in ['carrier_out', 'carrier_out_2', 'carrier_out_3']]):
            errors.append(
                '`{}` at `{}` is attempting to export a carrier '
                'not given as an output carrier: `{}`'.format(tech_id, loc_id, export)
            )

    # Error if non-allowed costs are defined
    for cost_class in tech_config.get_key('costs', {}):
        for k in tech_config.costs[cost_class]:
            if k not in allowed_costs:
                errors.append(
                    '`{}` at `{}` defines non-allowed '
                    '{} cost: `{}`'.format(tech_id, loc_id, cost_class, k)
                )
    return warnings, errors


def check_final(model_run):
    """
    Perform final checks of the completely built model_run.

    Returns
    -------
    comments : AttrDict
        debug output
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    warnings, errors = [], []
    comments = AttrDict()

    # Go through all loc-tech combinations and check validity
    for loc_id, loc_config in model_run.locations.items():
        if 'techs' in loc_config:
            for tech_id, tech_config in loc_config.techs.items():
                _check_tech(
                    model_run, tech_id, tech_config, loc_id,
                    warnings, errors, comments
                )

        if 'links' in loc_config:
            for link_id, link_config in loc_config.links.items():
                for tech_id, tech_config in link_config.techs.items():
                    _check_tech(
                        model_run, tech_id, tech_config,
                        'link {}:{}'.format(loc_id, link_id),
                        warnings, errors, comments
                    )
    # Either all locations or no locations must have coordinates
    all_locs = list(model_run.locations.keys())
    locs_with_coords = [
        k for k in model_run.locations.keys()
        if 'coordinates' in model_run.locations[k]
    ]
    if len(locs_with_coords) != 0 and len(all_locs) != len(locs_with_coords):
        errors.append(
            'Either all or no locations must have `coordinates` defined. '
            'Locations defined: {} - Locations with coordinates: {}'.format(
                all_locs, locs_with_coords)
        )

    # If locations have coordinates, they must all be either lat/lon or x/y
    elif len(locs_with_coords) != 0:
        first_loc = list(model_run.locations.keys())[0]
        try:
            coord_keys = sorted(list(model_run.locations[first_loc].coordinates.keys()))
            if coord_keys != ['lat', 'lon'] and coord_keys != ['x', 'y']:
                errors.append(
                    'Unidentified coordinate system. All locations must either'
                    'use the format {lat: N, lon: M} or {x: N, y: M}.'
                )
        except AttributeError:
            errors.append(
                'Coordinates must be given in the format {lat: N, lon: M} or '
                '{x: N, y: M}, not ' + str(model_run.locations[first_loc].coordinates)
            )

        for loc_id, loc_config in model_run.locations.items():
            try:
                if sorted(list(loc_config.coordinates.keys())) != coord_keys:
                    errors.append('All locations must use the same coordinate format.')
                    break
            except AttributeError:
                errors.append(
                    'Coordinates must be given in the format {lat: N, lon: M} or '
                    '{x: N, y: M}, not ' + str(model_run.locations[first_loc].coordinates)
                )
                break

    # Ensure that timeseries have no non-unique index values
    for k, df in model_run['timeseries_data'].items():
        if df.index.duplicated().any():
            errors.append('Time series `{}` contains non-unique timestamp values.'.format(k))

    # FIXME:
    # make sure `comments` is at the the base level:
    # i.e. comments.model_run.xxxxx....

    return comments, warnings, errors


def check_model_data(model_data):
    """
    Perform final checks of the completely built xarray Dataset `model_data`.

    Returns
    -------
    comments : AttrDict
        debug output
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    warnings, errors = [], []
    comments = AttrDict()

    # Ensure that no loc-tech specifies infinite resource and force_resource=True
    if "force_resource" in model_data.data_vars:
        relevant_loc_techs = [
            i.loc_techs_finite_resource.item()
            for i in model_data.force_resource if i.item() is True
        ]
        forced_resource = model_data.resource.loc[
            dict(loc_techs_finite_resource=relevant_loc_techs)
        ]
        conflict = forced_resource.where(forced_resource == np.inf).to_pandas().dropna()
        if not conflict.empty:
            errors.append(
                'loc_tech(s) {} cannot have `force_resource` set as infinite '
                'resource values are given'.format(', '.join(conflict.index))
            )

    # Ensure that if a tech has negative costs, there is a max cap defined
    # FIXME: doesn't consider capacity being set by a linked constraint e.g.
    # `resource_cap_per_energy_cap`.
    relevant_caps = [
        i for i in ['energy_cap', 'storage_cap', 'resource_cap', 'resource_area']
        if 'cost_' + i in model_data.data_vars.keys()
    ]
    for cap in relevant_caps:
        relevant_loc_techs = (model_data['cost_' + cap]
                              .where(model_data['cost_' + cap] < 0, drop=True)
                              .to_pandas())
        cap_max = cap + '_max'
        cap_equals = cap + '_equals'
        for loc_tech in relevant_loc_techs.columns:
            try:
                cap_val = model_data[cap_max][loc_tech].item()
            except KeyError:
                try:
                    cap_val = model_data[cap_equals][loc_tech].item()
                except KeyError:
                    cap_val = np.nan
            if np.isinf(cap_val) or np.isnan(cap_val):
                errors.append(
                    'loc_tech {} cannot have a negative cost_{} as the '
                    'corresponding capacity constraint is not set'
                    .format(loc_tech, cap)
                )

    for loc_tech in set(model_data.loc_techs_demand.values).intersection(model_data.loc_techs_finite_resource.values):
        if any(model_data.resource.loc[loc_tech].values > 0):
            errors.append(
                'Positive resource given for demand loc_tech {}. All demands '
                'must have negative resource'.format(loc_tech)
            )

    # Delete all empty dimensions & the variables associated with them
    for dim_name, dim_length in model_data.dims.items():
        if dim_length == 0:
            if dim_name in model_data.coords.keys():
                del model_data[dim_name]
            associated_vars = [
                var for var, data in model_data.data_vars.items() if dim_name in data.dims
            ]
            model_data = model_data.drop(associated_vars)
            warnings.append(
                'dimension {} and associated variables {} were empty, so have '
                'been deleted'.format(dim_name, ', '.join(associated_vars))
            )

    # Check if we're allowed to use operate mode
    if 'allow_operate_mode' not in model_data.attrs.keys():
        daily_timesteps = [
            model_data.timestep_resolution.loc[i].values
            for i in np.unique(model_data.timesteps.to_index().strftime('%Y-%m-%d'))
        ]
        if not np.all(daily_timesteps == daily_timesteps[0]):
            model_data.attrs['allow_operate_mode'] = 0
            warnings.append(
                'Operational mode requires the same timestep resolution profile '
                'to be emulated on each date'
            )
        else:
            model_data.attrs['allow_operate_mode'] = 1

    return model_data, comments, warnings, errors



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

constraint_sets.py
~~~~~~~~~~~~~~~~~~

loc_techs, loc_carriers, and loc_tech_carriers subsets used per constraint, to
reduce constraint complexity

"""

from calliope.core.preprocess.util import constraint_exists

import numpy as np

def generate_constraint_sets(model_run):
    """
    Generate loc-tech sets for a given pre-processed ``model_run``

    Parameters
    ----------
    model_run : AttrDict
    """

    sets = model_run.sets
    ## From here on, everything is a `key=value` pair within a dictionary

    constraint_sets = dict(
    # energy_balance.py
    loc_carriers_system_balance_constraint = sets.loc_carriers,
    loc_techs_balance_supply_constraint = sets.loc_techs_finite_resource_supply,
    loc_techs_balance_demand_constraint = sets.loc_techs_finite_resource_demand,
    loc_techs_resource_availability_supply_plus_constraint = sets.loc_techs_finite_resource_supply_plus,
    loc_techs_balance_transmission_constraint = sets.loc_techs_transmission,
    loc_techs_balance_supply_plus_constraint = sets.loc_techs_supply_plus,
    loc_techs_balance_storage_constraint = sets.loc_techs_storage,
    carriers_reserve_margin_constraint = [
        i for i in sets.carriers
        if i in model_run.model.get_key('reserve_margin', {}).keys()
    ],

    # costs.py
    loc_techs_cost_constraint = sets.loc_techs_cost,
    loc_techs_cost_investment_constraint = sets.loc_techs_investment_cost,
    loc_techs_cost_var_constraint = [
        i for i in sets.loc_techs_om_cost
        if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion
    ],

    # export.py
    loc_carriers_update_system_balance_constraint = [
        i for i in sets.loc_carriers if sets.loc_techs_export
        and any(['{0}::{2}'.format(*j.split('::')) == i
                for j in sets.loc_tech_carriers_export])
        ],
    loc_tech_carriers_export_balance_constraint = sets.loc_tech_carriers_export,
    loc_techs_update_costs_var_constraint = [
        i for i in sets.loc_techs_om_cost if i in sets.loc_techs_export
    ],
    loc_tech_carriers_export_max_constraint = [
        i for i in sets.loc_tech_carriers_export
        if constraint_exists(
            model_run, i.rsplit('::', 1)[0], 'constraints.export_cap'
        ) is not None
    ],

    # capacity.py
    loc_techs_storage_capacity_constraint = [
        i for i in sets.loc_techs_store if i not in sets.loc_techs_milp
    ],
    loc_techs_energy_capacity_storage_constraint = [
        i for i in sets.loc_techs_store
        if constraint_exists(model_run, i, 'constraints.charge_rate')
    ],
    loc_techs_resource_capacity_constraint = [
        i for i in sets.loc_techs_finite_resource_supply_plus
        if any([constraint_exists(model_run, i, 'constraints.resource_cap_equals'),
                constraint_exists(model_run, i, 'constraints.resource_cap_max'),
                constraint_exists(model_run, i, 'constraints.resource_cap_min')])
    ],
    loc_techs_resource_capacity_equals_energy_capacity_constraint = [
        i for i in sets.loc_techs_finite_resource_supply_plus
        if constraint_exists(model_run, i, 'constraints.resource_cap_equals_energy_cap')
    ],
    loc_techs_resource_area_constraint = [
        i for i in sets.loc_techs_area
    ],
    loc_techs_resource_area_per_energy_capacity_constraint = [
        i for i in sets.loc_techs_area
        if constraint_exists(model_run, i, 'constraints.resource_area_per_energy_cap')
        is not None
    ],
    locs_resource_area_capacity_per_loc_constraint = [
        i for i in sets.locs
        if model_run.locations[i].get_key('available_area', None) is not None
        and sets.loc_techs_area
    ],
    loc_techs_energy_capacity_constraint = [
        i for i in sets.loc_techs
        if i not in sets.loc_techs_milp + sets.loc_techs_purchase
    ],
    techs_energy_capacity_systemwide_constraint = [
        i for i in sets.techs
        if model_run.get_key('techs.{}.constraints.energy_cap_max_systemwide'.format(i), None)
        or model_run.get_key('techs.{}.constraints.energy_cap_equals_systemwide'.format(i), None)
    ],

    # dispatch.py
    loc_tech_carriers_carrier_production_max_constraint = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ],
    loc_tech_carriers_carrier_production_min_constraint = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ],
    loc_tech_carriers_carrier_consumption_max_constraint = [
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
    ],
    loc_techs_resource_max_constraint = sets.loc_techs_supply_plus,
    loc_techs_storage_max_constraint = sets.loc_techs_store,
    loc_tech_carriers_ramping_constraint = [
        i for i in sets.loc_tech_carriers_prod
        if i.rsplit('::', 1)[0] in sets.loc_techs_ramping
    ],

    # milp.py
    loc_techs_unit_commitment_constraint = sets.loc_techs_milp,
    loc_techs_unit_capacity_constraint = sets.loc_techs_milp,
    loc_tech_carriers_carrier_production_max_milp_constraint = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ],
    loc_techs_carrier_production_max_conversion_plus_milp_constraint = [
        i for i in sets.loc_techs_conversion_plus
        if i in sets.loc_techs_milp
    ],
    loc_tech_carriers_carrier_production_min_milp_constraint = [
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ],
    loc_techs_carrier_production_min_conversion_plus_milp_constraint = [
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i in sets.loc_techs_milp
    ],
    loc_tech_carriers_carrier_consumption_max_milp_constraint = [
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
    ],
    loc_techs_energy_capacity_units_constraint = [
        i for i in sets.loc_techs_milp
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_unit')
        is not None
    ],
    loc_techs_storage_capacity_units_constraint = [
        i for i in sets.loc_techs_milp if i in sets.loc_techs_store
    ],
    loc_techs_energy_capacity_max_purchase_constraint = [
        i for i in sets.loc_techs_purchase
        if (constraint_exists(model_run, i, 'constraints.energy_cap_equals') is not None
            or (constraint_exists(model_run, i, 'constraints.energy_cap_max') is not None
                and constraint_exists(model_run, i, 'constraints.energy_cap_max') != np.inf))
    ],
    loc_techs_energy_capacity_min_purchase_constraint = [
        i for i in sets.loc_techs_purchase
        if (not constraint_exists(model_run, i, 'constraints.energy_cap_equals')
            and constraint_exists(model_run, i, 'constraints.energy_cap_min'))
    ],
    loc_techs_storage_capacity_max_purchase_constraint = [
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (constraint_exists(model_run, i, 'constraints.storage_cap_equals') is not None
            or (constraint_exists(model_run, i, 'constraints.storage_cap_max') is not None
                and constraint_exists(model_run, i, 'constraints.storage_cap_max') != np.inf))
    ],
    loc_techs_storage_capacity_min_purchase_constraint = [
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (not constraint_exists(model_run, i, 'constraints.storage_cap_equals')
            and constraint_exists(model_run, i, 'constraints.storage_cap_min'))
    ],
    loc_techs_update_costs_investment_units_constraint = [
        i for i in sets.loc_techs_milp
        if i in sets.loc_techs_investment_cost and
        any(constraint_exists(model_run, i, 'costs.{}.purchase'.format(j))
               for j in model_run.sets.costs)
    ],
    # loc_techs_purchase technologies only exist becuase they have defined a purchase cost
    loc_techs_update_costs_investment_purchase_constraint = sets.loc_techs_purchase,

    # conversion.py
    loc_techs_balance_conversion_constraint = sets.loc_techs_conversion,
    loc_techs_cost_var_conversion_constraint = sets.loc_techs_om_cost_conversion,

    # conversion_plus.py
    loc_techs_balance_conversion_plus_primary_constraint = sets.loc_techs_conversion_plus,
    loc_techs_carrier_production_max_conversion_plus_constraint = [
        i for i in sets.loc_techs_conversion_plus
        if i not in sets.loc_techs_milp
    ],
    loc_techs_carrier_production_min_conversion_plus_constraint = [
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i not in sets.loc_techs_milp
    ],
    loc_techs_cost_var_conversion_plus_constraint = sets.loc_techs_om_cost_conversion_plus,
    loc_techs_balance_conversion_plus_in_2_constraint = sets.loc_techs_in_2,
    loc_techs_balance_conversion_plus_in_3_constraint = sets.loc_techs_in_3,
    loc_techs_balance_conversion_plus_out_2_constraint = sets.loc_techs_out_2,
    loc_techs_balance_conversion_plus_out_3_constraint = sets.loc_techs_out_3,

    # network.py
    loc_techs_symmetric_transmission_constraint = sets.loc_techs_transmission,

    # policy.py
    techlists_group_share_energy_cap_min_constraint = [
        i for i in sets.techlists
        if 'energy_cap_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ],
    techlists_group_share_energy_cap_max_constraint = [
        i for i in sets.techlists
        if 'energy_cap_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ],
    techlists_group_share_energy_cap_equals_constraint = [
        i for i in sets.techlists
        if 'energy_cap_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
    ],
    techlists_carrier_group_share_carrier_prod_min_constraint = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_min'.format(i), {}).keys()
    ],
    techlists_carrier_group_share_carrier_prod_max_constraint = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_max'.format(i), {}).keys()
    ],
    techlists_carrier_group_share_carrier_prod_equals_constraint = [
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_equals'.format(i), {}).keys()
    ],

    ) ## End of dictionary

    return constraint_sets
